// BufferingFilteredImageSource class. 

import java.awt.image.*;
import java.util.Hashtable;
import java.util.Enumeration;

public class BufferingFilteredImageSource implements ImageProducer {
    protected int width, height;
    protected int[] pixels = null;
    protected ImageProducer source = null;
    protected boolean done = false;
    protected boolean isSingleFrame = false;
    
    public BufferingFilteredImageSource( ImageProducer orig, ImageFilter imgf) {        
        // send the origianl image producer pixel data through the filter...
        source = new FilteredImageSource( orig, imgf);

        // this block of code is synchronized with the imageComplete
        // function of the internal consumer object to ensure that the 
        // image is completely filtered and stored in the pixels array 
        // before we construct the internal in-memory image source.
        //
        // we need the boolean flag as well as the wait() call because
        // the call to startProduction may return immediately and deliver
        // the image on another thread, or it may deliver the image
        // using the calling thread.
        BfisInternalConsumer bfisic = new BfisInternalConsumer();
        synchronized(bfisic) {
            // deliver the filtered image to the internal consumer...
            source.startProduction(bfisic);
            if (!done) {
                try {
                    bfisic.wait();
                } catch (Exception e) {}
            }
        }

        // create an in-memory image source from the samples stored in
        // the pixels array if the image is a single frame.
        // if the image is a sequence of images, use the FilteredImageSource
        // already created as the internal source...
        if (isSingleFrame) {
            source = new MemoryImageSource( width, height, pixels, 0, width);
        }

        // release our reference to the pixels buffer...
        pixels = null;
    }
    
    public void addConsumer(ImageConsumer ic) {
        // forward to the ImageProducer member object...
        source.addConsumer(ic);
    }

    public boolean isConsumer(ImageConsumer ic) {
        // forward to the ImageProducer member object...
        return source.isConsumer(ic);
    }
    public void removeConsumer(ImageConsumer ic) {
        // forward to the ImageProducer member object...
        source.removeConsumer(ic);
    }

    public void requestTopDownLeftRightResend(ImageConsumer ic) {
        // forward to the ImageProducer member object...
        source.requestTopDownLeftRightResend(ic);
    }

    public void startProduction(ImageConsumer ic) {
        // forward to the ImageProducer member object...
        source.startProduction(ic);
    }

    // this inner class consumes the output of the FilteredImageSource
    // and stores it in the pixels array in the containing 
    // BufferingFilteredImageSource object...
    private class BfisInternalConsumer implements ImageConsumer {
        public void imageComplete(int status) {
            if (!done) {
                // this is the first frame so determine if the image
                // is single frame or multi-frame and notify the 
                // BufferingFilteredImageSource containing object in
                // case it is waiting...                
                synchronized(this) {
                    BufferingFilteredImageSource.this.isSingleFrame = ((status & STATICIMAGEDONE) == STATICIMAGEDONE);
                    BufferingFilteredImageSource.this.done = true;
                    notify();

                    // remove the internal consumer from the source client list...
                    source.removeConsumer(this);                    
                }
            }
        }

        public void setColorModel(ColorModel cm) {
            // do nothing...
            return;
        }

        public void setDimensions( int width, int height) {
            // save the image dimensions and allocate a buffer
            // using the member variables of the containing class...
            BufferingFilteredImageSource.this.width = width;
            BufferingFilteredImageSource.this.height = height;
            BufferingFilteredImageSource.this.pixels = new int[width*height];
        }

        public void setHints( int hints) {
            // do nothing...
            return;
        }

        public void setPixels( int x, int y, int w, int h, ColorModel cm, byte[] pixels, int offset, int scansize) {
            // ensure that the pixel buffer has been allocated...
            if (BufferingFilteredImageSource.this.pixels == null) {
                return;
            }
            
            // loop through all the samples in the pixels buffer and place them
            // in the proper locations in the color component buffers...
            int srcpos = offset;
            int destpos = y*BufferingFilteredImageSource.this.width + x;
            for (int j = 0; j < h; j++) {
                for (int i = 0; i < w; i++) {
                    // use the color model parameter to extract each
                    // of the 4 components...
                    // the pixel value must be converted to an integer
                    // between 0 ang 255, we need to clear the upper
                    // bits to zero because java integer types are signed
                    // and the cast from a byte to an int will result in
                    // negative numbers for byte values greater than or
                    // equal to 0x80...
                    int t = ((int)pixels[srcpos + i] & 0x000000FF);

                    // use the color model parameter to extract each
                    // of the 4 components...
                    int r = cm.getRed(t);
                    int g = cm.getGreen(t);
                    int b = cm.getBlue(t);
                    int a = cm.getAlpha(t);
                    BufferingFilteredImageSource.this.pixels[destpos + i] = (a << 24) | (r << 16) | (g << 8) | b;
                }
                srcpos += scansize;
                destpos += width;
            }
        }

        public void setPixels( int x, int y, int w, int h, ColorModel cm, int[] pixels, int offset, int scansize) {
            // ensure that the pixel buffer has been allocated...
            if (BufferingFilteredImageSource.this.pixels == null) {
                return;
            }
            
            // loop through all the samples in the pixels buffer and place them
            // in the proper locations in the color component buffers...
            int srcpos = offset;
            int destpos = y*BufferingFilteredImageSource.this.width + x;
            for (int j = 0; j < h; j++) {
                for (int i = 0; i < w; i++) {
                    // use the color model parameter to extract each
                    // of the 4 components...
                    int r = cm.getRed(pixels[srcpos + i]);
                    int g = cm.getGreen(pixels[srcpos + i]);
                    int b = cm.getBlue(pixels[srcpos + i]);
                    int a = cm.getAlpha(pixels[srcpos + i]);
                    BufferingFilteredImageSource.this.pixels[destpos + i] = (a << 24) | (r << 16) | (g << 8) | b;
                }
                srcpos += scansize;
                destpos += width;
            }
        }

        public void setProperties( Hashtable props) {
            // do nothing...
            return;
        }
    }
}

// The JIPTestApplet. 
import java.applet.*;
import java.awt.*;
import java.awt.image.*;
import java.net.URL;

public class JIPTestApplet extends Applet {
    // program constants
    protected static final int layoutGap = 5;
    protected static final int maxImages = 10;

    // data members...
    protected Button greyOutButton;
    protected Button rotateClockwiseButton;
    protected Button sharpenButton;
    protected Button revertButton;
    protected ImageCanvas imageCanvas;
    protected Image currentImage;
    protected Checkbox[] filteringMethodRadioButtons = new Checkbox[2];
    protected Checkbox[] imageRadioButtons = new Checkbox[maxImages];
    protected Image[] originalImages = new Image[maxImages];
    protected int imageCount = 0;
    protected int currentImageIndex = 0;
    protected Color appletColor = Color.lightGray;
    protected Color canvasColor = Color.darkGray;

    public void init() {
        createUserInterface();
    }

    public void destroy() {
        // release resources...
        greyOutButton = null;
        rotateClockwiseButton = null;
        sharpenButton = null;
        revertButton = null;
        imageCanvas = null;
        currentImage = null;
        imageRadioButtons = null;
        filteringMethodRadioButtons = null;
        originalImages = null;
    }

    private void createUserInterface() {
        // set the applet background...
        setBackground(appletColor);

        // initialize filter button panel...
        Panel filterButtonPanel = new OutlinedPanel();
        filterButtonPanel.setLayout( new FlowLayout( FlowLayout.CENTER, layoutGap, layoutGap));
        filterButtonPanel.add( greyOutButton = new Button("Grey Out"));
        filterButtonPanel.add( rotateClockwiseButton = new Button("Rotate Clockwise"));
        filterButtonPanel.add( sharpenButton = new Button("Sharpen"));
        filterButtonPanel.add( revertButton = new Button("Revert"));

        // initialize filtering method button panel...
        Panel filteringMethodButtonPanel = new OutlinedPanel();
        CheckboxGroup fmcbg = new CheckboxGroup();
        filteringMethodButtonPanel.setLayout( new FlowLayout( FlowLayout.CENTER, layoutGap, layoutGap));
        filteringMethodButtonPanel.add( filteringMethodRadioButtons[0] = new Checkbox( "FilteredImageSource", fmcbg, true));
        filteringMethodButtonPanel.add( filteringMethodRadioButtons[1] = new Checkbox( "BufferingFilteredImageSource", fmcbg, false));

        // create the image button panel...
        Panel imageButtonPanel = new OutlinedPanel();
        imageButtonPanel.setLayout( new FlowLayout( FlowLayout.CENTER, layoutGap, layoutGap));

        // load the images and add the image radio
        // buttons to the image button panel...
        CheckboxGroup cbg = new CheckboxGroup();
        for (int i = 0; i < maxImages; i++) {
            String filename = getParameter("image" + (i + 1));
            if (filename != null) {
                try {
                    URL url = new URL( getDocumentBase(), filename);
                    originalImages[imageCount] = getImage(url);
                    int nameStart = filename.lastIndexOf('/') + 1;
                    Checkbox cb = new Checkbox( filename.substring(nameStart), cbg, (imageCount == 0) ? true : false);
                    imageButtonPanel.add( imageRadioButtons[imageCount] = cb);
                    imageCount++;
                }
                catch ( Exception e) {
                }
            }
            else {
                break;
            }
        }

        // initialize layout...
        setLayout(new FlowLayout( FlowLayout.CENTER, layoutGap, layoutGap));
        add(imageCanvas = new ImageCanvas());
        add(imageButtonPanel);
        add(filteringMethodButtonPanel);
        add(filterButtonPanel);

        // compute canvas size and resize canvas...
        Dimension d = size();
        Dimension dp = filterButtonPanel.preferredSize();
        imageCanvas.setBackground( canvasColor);
        imageCanvas.resize( d.width - 2*layoutGap, d.height - 2*dp.height - 12*layoutGap);

        // set the current image...
        currentImage = originalImages[currentImageIndex];

        // repaint the GUI...
        repaint();
    }

    public void paint( Graphics g) {
        // now that the update() call has filled the Applet panel
        // draw a 3D outline around the panel one pixel in from the edges...
        Dimension d = size();
        g.setColor( getBackground());
        g.draw3DRect( 1, 1, d.width - 2, d.height - 2, false);
        super.paint(g);
    }

    public boolean action( Event e, Object arg) {
        // create the filter based on the event target...
        ImageFilter filter = null;
        if (e.target == greyOutButton) {
            filter = new GreyOutImageFilter();
        }
        else if (e.target == rotateClockwiseButton) {
            filter = new RotateClockwiseImageFilter();
        }
        else if (e.target == sharpenButton) {
            filter = new SharpenImageFilter();
        }
        else if (e.target == revertButton) {
            // not filtering, just restore the
            // original image...
            currentImage = originalImages[currentImageIndex];
        }
        else {
            // check if the target is one of the image buttons...
            boolean found = false;
            for (int i = 0; i < imageCount; i++) {
                if (e.target == imageRadioButtons[i]) {
                    found = true;
                    currentImage = originalImages[currentImageIndex = i];
                    break;
                }
            }

            // otherwise return false since we didn't handle
            // action event...
            if (found == false)
                return false;
        }

        // apply the filter if any with the currently selected filtering method...
        if (filter != null) {
            if (filteringMethodRadioButtons[0].getState())
                currentImage = createImage( new FilteredImageSource( currentImage.getSource(), filter));
            else
                currentImage = createImage( new BufferingFilteredImageSource( currentImage.getSource(), filter));
        }

        // display the result...
        imageCanvas.repaint();
        
        return true;
    }

    private class ImageCanvas extends Canvas {
        public void paint( Graphics g) {
            // display the image centered in the canvas...
            if (currentImage != null) {
                Dimension d = size();
                int w = currentImage.getWidth(this);
                int h = currentImage.getHeight(this);
                if ((w > 0) && (h > 0)) {
                    // paint the image if we know the dimensions...
                    g.drawImage( currentImage, (d.width - w) / 2, (d.height - h) / 2, this);
                }
                else {
                    // otherwise the default imageUpdate handler will
                    // call repaint() when the image is available...
                    prepareImage( currentImage, this);
                }
            }
        }
        
        public boolean imageUpdate( Image img, int flags, int x, int y, int width, int height) {
            // we only care about information for the current image...
            if (img != currentImage)
                return false;
            else
                return super.imageUpdate( img, flags, x, y, width, height);
        }            
    }

    private class OutlinedPanel extends Panel {
        public void paint( Graphics g) {
            // now that the update() call has filled the panel
            // with the background color, all we need to do is
            // draw the 3D outline around the panel one
            // pixel in from the edges...
            Dimension d = size();
            g.setColor( getBackground());
            g.draw3DRect( 1, 1, d.width - 2, d.height - 2, false);
        }
    }
}

13


