// DIBDoc.cpp : implementation file
//

#include "stdafx.h"
#include "Segment.h"
#include "DIBDoc.h"
#include "MonochromeBitmap.h"
#include "Line.h"
#include "HistogramDlg.h"
#include <vector>
#include <algorithm>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//-----------------------------------------------------------------
//
// CDIBMonochromeBitmap - this forms the base class of the monochrome
//						  bitmaps that extract their rasters from
//						  device independent bitmaps
//
//-----------------------------------------------------------------

class CDIBMonochromeBitmap: public CMonochromeBitmap
{
protected:
	CDIBDoc *m_pDoc;
public:
	CDIBMonochromeBitmap() { m_pDoc = NULL; }
	void SetDoc(CDIBDoc &doc) {
		m_pDoc = &doc;
	}

	virtual int Width() const {
		ASSERT(m_pDoc);
		return m_pDoc->GetBitmapInfo()->bmiHeader.biWidth;
	}

	virtual int Height() const {
		ASSERT(m_pDoc);
		return m_pDoc->GetBitmapInfo()->bmiHeader.biHeight;
	}
};

class CDIB1MonochromeBitmap: public CDIBMonochromeBitmap
{
public:
	const BYTE *GetRaster(int nLine);
};

const BYTE *CDIB1MonochromeBitmap::GetRaster(int nLine)
{
	ASSERT(m_pDoc);
	return m_pDoc->GetRaster(nLine);
}

//-----------------------------------------------------------------
//
// CDIB4MonochromeBitmap - this is a monochrome bitmap that extracts
//						   its rasters from a 16 color DIB. 
//
// NOTE: This class really hasn't been tested.
//-----------------------------------------------------------------

class CDIB4MonochromeBitmap: public CDIBMonochromeBitmap
{
public:
	CDIB4MonochromeBitmap();
	virtual ~CDIB4MonochromeBitmap();

	const BYTE *GetRaster(int nLine);
protected:
	BYTE *m_pRaster;
};

CDIB4MonochromeBitmap::CDIB4MonochromeBitmap()
{
	m_pRaster = NULL;
}

CDIB4MonochromeBitmap::~CDIB4MonochromeBitmap()
{
	delete [] m_pRaster;
}

const BYTE *CDIB4MonochromeBitmap::GetRaster(int nLine)
{
	if (m_pRaster == NULL) {
		ASSERT(m_pDoc);
		m_pRaster = new BYTE [m_pDoc->GetRasterWidthInBytes()];
	}


	const CByteArray &bThreshold = m_pDoc->GetThresholdTable();
	int nLimit = Width();

	const BYTE *pSrc = m_pDoc->GetRaster(nLine);
	BYTE *pDest = m_pRaster;
	for (int i = 0; i < nLimit; i += 8) {
		BYTE b = 0;
		if (bThreshold[*pSrc & 15]) {
			b = 0x80;
		}
		if (bThreshold[*pSrc++ >> 4]) {
			b += 0x40;
		}
		if (bThreshold[*pSrc & 15]) {
			b += 0x20;
		}
		if (bThreshold[*pSrc++ >> 4]) {
			b += 0x10;
		}
		if (bThreshold[*pSrc & 15]) {
			b += 0x08;
		}
		if (bThreshold[*pSrc++ >> 4]) {
			b += 0x04;
		}
		if (bThreshold[*pSrc & 15]) {
			b += 0x02;
		}
		if (bThreshold[*pSrc++ >> 4]) {
			b += 0x01;
		}
		*pDest++ = b;
	}
	return m_pRaster;
}

//-----------------------------------------------------------------
//
// CDIB8MonochromeBitmap - this one is the 256 color bitmap, the one
//					most frequently used to store pictures and gray
//					scales.
//
//-----------------------------------------------------------------

class CDIB8MonochromeBitmap: public CDIBMonochromeBitmap
{
public:
	virtual void GetBrightnessHistogram(vector<int> &a,vector<int> &vBrightness);
	virtual  ~CDIB8MonochromeBitmap();
	 CDIB8MonochromeBitmap();
	virtual const BYTE *GetRaster(int nLine);
	virtual void BuildBlobList();
protected:
	CByteArray m_bCachedRaster;
	BYTE * m_pCachedRaster;
};

//-----------------------------------------------------------------
//
// CDIB8MonochromeBitmap::CDIB8MonochromeBitmap
//
//-----------------------------------------------------------------

CDIB8MonochromeBitmap::CDIB8MonochromeBitmap()
{
}

//-----------------------------------------------------------------
//
// CDIB8MonochromeBitmap::~CDIB8MonochromeBitmap
//
//-----------------------------------------------------------------

CDIB8MonochromeBitmap::~CDIB8MonochromeBitmap()
{
}

//-----------------------------------------------------------------
//
// CDIB8MonochromeBitmap::GetBrightnessHistogram - This creates
//					a histogram whose elements indicate the relative
//				    number of pixels at each brightness. You can use
//					this to determine where to threshold.
//
// a - Histogram of # of pixels at each index. Indices are in order of
//     lowest to highest brightness.
// aBrightness - Each element of this array contains the index into
//				 the histogram of the indexed element.
//-----------------------------------------------------------------

void CDIB8MonochromeBitmap::GetBrightnessHistogram(vector<int> & a,vector<int> &aBrightness)
{
	// Here, we sample a bunch of random points to get an idea
	// of the brightness distribution.

	int nColors = m_pDoc->GetBitmapInfo()->bmiHeader.biClrUsed;
	if (nColors == 0) nColors = 256;
	RGBQUAD *pColors = m_pDoc->GetBitmapInfo()->bmiColors;

	a.resize(nColors);
	aBrightness.resize(nColors);
	for (int i = 0; i < nColors; i++) a[i] = 0;

	// We do a little trick here. We put the brightness in
	// bits 8+ of an int and the index in the low 7 bits.
	// We sort the array and we get the indices ordered by
	// brightness.

	vector<int> vBrightness;
	vBrightness.resize(nColors);
	for (i = 0; i < nColors; i++) {
		int nBrightness = pColors[i].rgbRed + pColors[i].rgbBlue + pColors[i].rgbGreen;
		vBrightness[i] = i + (nBrightness << 8);
	}

	sort(vBrightness.begin(), &vBrightness[vBrightness.size()]);

	// Now, we put the index in the top byte and put the ranking in the
	// bottom and sort. The "i"th index will contain the ranking of
	// the indexed color.

	for (i = 0; i < nColors; i++) {
		vBrightness[i] = ((vBrightness[i] & 0xFF) << 8) + i;
	}
	sort(vBrightness.begin(), &vBrightness[vBrightness.size()]);

	for (i = 0; i < nColors; i++) {
		vBrightness[i] &= 0xFF;
	}

	// Let's record the brightnesses in "aBrightness"

	for (i = 0; i < nColors; i++) {
		RGBQUAD *pClr = pColors + vBrightness[i];
		aBrightness[i] = pClr->rgbRed + pClr->rgbBlue + pClr->rgbGreen;
	}

	const int nWidth = Width();
	const int nHeight = Height();
	
	for (i = 0; i < 100; i++) {
		int nY = MulDiv(i,nHeight - 1,100);
		for (int j = 0; j < 100; j++) {
			int nX = MulDiv(j,nWidth - 1,100);
			BYTE bValue = m_pDoc->GetRaster(nY)[nX];
			ASSERT(vBrightness[bValue] >= 0);
			ASSERT(vBrightness[bValue] < nColors);
			a[vBrightness[bValue]]++;
		}
	}
}

//-----------------------------------------------------------------
//
// CDIB8MonochromeBitmap::GetRaster - return a pointer to a thresholded
//					monochrome raster (1 bit per pixel) for the indexed
//					line. NOTE: only one line is cached at a time, sorry.
//
// nLine - Y index of bitmap line to fetch.
//-----------------------------------------------------------------

const BYTE *CDIB8MonochromeBitmap::GetRaster(int nLine)
{
	if (m_bCachedRaster.GetSize() < Width()) {
		int nMonochromeWidth = m_pDoc->GetRasterWidthInBytes();
		m_bCachedRaster.SetSize(nMonochromeWidth);
		m_pCachedRaster = m_bCachedRaster.GetData();
		memset(m_pCachedRaster,0,nMonochromeWidth);
	}

	const CByteArray &m_bThreshold = m_pDoc->GetThresholdTable();

	BYTE *pDest = m_pCachedRaster;
	const BYTE *pSrc = m_pDoc->GetRaster(nLine);

	// We do 8 bit chunks here. We'll handle the end boundary
	// condition separately

	const int nWidth = Width() & ~ 7;
	BYTE b;
	for (int i = 0; i < nWidth; i += 8) {
		if (m_bThreshold[*pSrc++]) b = 0x80; else b = 0;
		if (m_bThreshold[*pSrc++]) b += 0x40;
		if (m_bThreshold[*pSrc++]) b += 0x20;
		if (m_bThreshold[*pSrc++]) b += 0x10;
		if (m_bThreshold[*pSrc++]) b += 0x08;
		if (m_bThreshold[*pSrc++]) b += 0x04;
		if (m_bThreshold[*pSrc++]) b += 0x02;
		if (m_bThreshold[*pSrc++]) b += 0x01;
		*pDest++ = b;
	}

	if (nWidth != Width()) {
		b = 0;
		BYTE bBit = 0x80;
		while(i++ < Width()) {
			if (m_bThreshold[*pSrc++]) b += bBit;
			bBit = bBit >> 1;
		}
		*pDest = b;
	}
	return m_pCachedRaster;
}

//-----------------------------------------------------------------
//
// CDIB8MonochromeBitmap::BuildBlobList - I've just duplicated the code
//									  pretty much for the x8 bitmap
//									  in order to make it go fast.
//
//-----------------------------------------------------------------

void CDIB8MonochromeBitmap::BuildBlobList()
{
	const CByteArray &bThresholdTable = m_pDoc->GetThresholdTable();

	CLineVectorIterator clviNew,clviEnd;
	AllocateLineBlock(clviNew,clviEnd);
	m_pBlob.reserve(eBlobsPerBlock);

	// Create two vectors of pointers to lines to hold the new line list
	// and the old line list. We initialize them to hold Width()/2 members.
	// This is the maximum possible number of lines.

	CLinePointerVector clpvRaster[2];
	clpvRaster[0].reserve(Width()/2);
	clpvRaster[1].reserve(Width()/2);

	int nOldLineIndex = 0;
	int nNewLineIndex = 1;

	// At the start, the old line list has no members.
	CLinePointerVectorIterator clpviOldEnd = clpvRaster[nOldLineIndex].begin();

	for (int nY = 0; nY < Height(); nY++) {

		CLinePointerVector &clpvOld = clpvRaster[nOldLineIndex];
		CLinePointerVector &clpvNew = clpvRaster[nNewLineIndex];
		nOldLineIndex = nOldLineIndex ^ 1;
		nNewLineIndex = nNewLineIndex ^ 1;

		CLinePointerVectorIterator clpviOld = clpvOld.begin();
		CLine *pLineOld;
		if (clpviOld == clpviOldEnd) {
			pLineOld = NULL;
		} else {
			pLineOld = *clpviOld++;
		}

		CLinePointerVectorIterator clpviNew = clpvNew.begin();

		// Get the raster to process and set up for the first byte
		const BYTE *pRaster = m_pDoc->GetRaster(nY);

		//  Set up the X extents.
		int nX = 0;
		const int nXEnd = Width();

		while(nX < nXEnd) {
			CBlob *pBlobOld;
			// Search for the start of a line.

			while(bThresholdTable[*pRaster++] == 0) {
				if (++nX == nXEnd) goto linedone;
			}

			// Start a line.

			if (clviNew == clviEnd) {
				// need more memory for lines.

				AllocateLineBlock(clviNew,clviEnd);
			}

			CLine &lineNew = *clviNew++;
			lineNew.m_nXStart = nX;
			lineNew.m_pNextLineSameBlob = 0;
			lineNew.m_nY = nY;
			
			// Put it on the new list.

			*clpviNew++ = &lineNew;

			// Find the extent of the line.

			nX++;
			while (nX != nXEnd) {
				if (bThresholdTable[*pRaster++] == 0) break;
				nX++;
			}

			lineNew.m_nXEnd = nX++;

			// Now finish all old lines wholly before our new line

			if (pLineOld) {
				while (pLineOld->m_nXEnd < lineNew.m_nXStart) {

					pBlobOld = pLineOld->m_pBlob;
					if (--(pBlobOld->m_nOpenLines) == 0) {
						AddBlob(pBlobOld);
					}
					if (clpviOld == clpviOldEnd) {
						pLineOld = NULL;
						break;
					} else {
						pLineOld = *clpviOld++;
					}
				}
			}

			// Do the first line that overlaps our new line

			if (pLineOld && pLineOld->m_nXStart <= lineNew.m_nXEnd) {
				pBlobOld = pLineOld->m_pBlob;
				lineNew.m_pBlob = pBlobOld;
				*(pBlobOld->m_ppLastLine) = &lineNew;
				pBlobOld->m_ppLastLine = &(lineNew.m_pNextLineSameBlob);

				// See if the new line ends the old line or vice-versa.

				if (pLineOld->m_nXEnd > lineNew.m_nXEnd) {
					// the old line extends past the new line.
					// the new line gives the old line's blob another open line.

					pBlobOld->m_nOpenLines++;
					continue;
				} else {
					// we close the old line and extend the blob at the
					// same time. We continue to close old lines.

					if (clpviOld == clpviOldEnd) {
						pLineOld = NULL;
					} else {
						pLineOld = *clpviOld++;
						while(pLineOld->m_nXEnd <= lineNew.m_nXEnd) {
							// End this line too. We have two cases:
							CBlob *pBlobOther;
							if ((pBlobOther = pLineOld->m_pBlob) == pBlobOld) {
								// the old line is part of the new line's blob.
								// This is just a loop.

								pBlobOld->m_nLoops++;
								pBlobOld->m_nOpenLines--;
							} else {
								// This is the merge case

								pBlobOld->Merge(pBlobOther);
								delete pBlobOther;
								pBlobOld->m_nOpenLines--;
							}
							if (clpviOld == clpviOldEnd) {
								pLineOld = NULL;
								break;
							} else {
								pLineOld = *clpviOld++;
							}
						}
						if (pLineOld && pLineOld->m_nXStart <= lineNew.m_nXEnd) {
							// This last old line overlaps and ends the new line.

							CBlob *pBlobOther = pLineOld->m_pBlob;
							if (pBlobOther == pBlobOld) {
								pBlobOld->m_nLoops++;
							} else {
								pBlobOld->Merge(pBlobOther);
								delete pBlobOther;
							}
						}
					}
				}
			} else {
				// This is the case where no line overlaps the new line.
				// Start a blob with this line.

				CBlob *pBlobNew = lineNew.m_pBlob = new CBlob;
				pBlobNew->m_nOpenLines = 1;
				pBlobNew->m_pFirstLine = &lineNew;
				pBlobNew->m_ppLastLine = &(lineNew.m_pNextLineSameBlob);
				pBlobNew->m_nLoops = 0;
			}
		}
linedone:
		if (pLineOld) {
			while (TRUE) {
				// Finish all old lines.
				CBlob *pBlobOld = pLineOld->m_pBlob;
				if (--(pBlobOld->m_nOpenLines) == 0) {
					AddBlob(pBlobOld);
				}
				if (clpviOld == clpviOldEnd) break;
				pLineOld = *clpviOld++;
			}
		}

		// Finally, record the end of the new list as the
		// end of the old list to be.

		clpviOldEnd = clpviNew;
	}

	// At the end, we've got one last raster of lines to finish.

	CLinePointerVector &clpvLast = clpvRaster[nOldLineIndex];
	for (CLinePointerVectorIterator clpvi = clpvLast.begin();
	clpvi != clpviOldEnd;) {
		// Finish all old lines.
		CBlob *pBlobOld = (*clpvi++)->m_pBlob;
		if (--(pBlobOld->m_nOpenLines) == 0) {
			AddBlob(pBlobOld);
		}
	}

	// and finally, segmentation is complete
}

/////////////////////////////////////////////////////////////////////////////
// CDIBDoc - This implements CImageDoc for a device independent bitmap.

IMPLEMENT_DYNCREATE(CDIBDoc, CImageDoc)

//-----------------------------------------------------------------
//
// CDIBDoc::CDIBDoc - constructor
//
//-----------------------------------------------------------------

CDIBDoc::CDIBDoc()
{
	m_pMonochromeBitmap = NULL;
	m_hFileMapping = NULL;
	m_pFile = NULL;
	m_pBMFH = NULL;
	m_pBMI = NULL;
	m_eThresholdMethod = eWhiteBackground;
	m_nThreshold = 128*3;
}

//-----------------------------------------------------------------
//
// CDIBDoc::OnNewDocument - This resets the doc when it gets reused.
//
//-----------------------------------------------------------------

BOOL CDIBDoc::OnNewDocument()
{
	if (! CImageDoc::OnNewDocument()) return FALSE;

	if (m_file.m_hFile != CFile::hFileNull) {
		CloseFile();
		ASSERT(m_file.m_hFile == CFile::hFileNull);
	}
	return TRUE;
}

CDIBDoc::~CDIBDoc()
{
	if (! IsEmpty())
		CloseFile();	
	if (m_pMonochromeBitmap) delete m_pMonochromeBitmap;
}


BEGIN_MESSAGE_MAP(CDIBDoc, CImageDoc)
	//{{AFX_MSG_MAP(CDIBDoc)
	ON_COMMAND(ID_IMAGE_HISTOGRAM, OnImageHistogram)
	ON_UPDATE_COMMAND_UI(ID_IMAGE_HISTOGRAM, OnUpdateImageHistogram)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDIBDoc diagnostics

#ifdef _DEBUG
void CDIBDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CDIBDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CDIBDoc commands

//-----------------------------------------------------------------
//
// CDIBDoc::OnOpenDocument - Open a DIB directly. Don't go through
//							 MFC serialize. We map the file into
//							 memory and let its consumers hack it.
//
// lpszPathName - open this file.
//
//-----------------------------------------------------------------

BOOL CDIBDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
	if (!CDocument::OnOpenDocument(lpszPathName))
		return FALSE;
	
	if (m_file.m_hFile != CFile::hFileNull) CloseFile();
	
	CFileException e;
	if (! m_file.Open(lpszPathName,CFile::modeRead|CFile::shareDenyWrite,&e)) {
		AfxThrowFileException(e.m_cause,e.m_lOsError,lpszPathName);
	}
	
	m_hFileMapping = CreateFileMapping((HANDLE)(m_file.m_hFile),NULL,PAGE_READONLY,0,0,NULL);
	if (m_hFileMapping == NULL) {
		m_file.Close();
		CFileException::ThrowOsError(GetLastError(),lpszPathName);
	}

	m_pFile = MapViewOfFile(m_hFileMapping,FILE_MAP_READ,0,0,0);
	if (m_pFile == NULL) {
		m_file.Close();
		CFileException::ThrowOsError(GetLastError(),lpszPathName);
	}

	// Here's a bunch of useful pointers.

	m_pBMFH = (BITMAPFILEHEADER *)m_pFile;
	if (m_pBMFH->bfType != 0x4d42) {
		CString csz;
		AfxFormatString1(csz,IDS_FILE_TYPE_ERROR,lpszPathName);
		AfxMessageBox(csz);
		return FALSE;
	}

	m_pBMI = (BITMAPINFO *)(m_pBMFH + 1);
	if (m_pBMI->bmiHeader.biSize != sizeof(BITMAPINFOHEADER) ||
		m_pBMI->bmiHeader.biCompression != BI_RGB ||
		m_pBMI->bmiHeader.biPlanes != 1) {
		CString csz;
		AfxFormatString1(csz,IDS_FILE_FORMAT_NOT_SUPPORTED,lpszPathName);
		AfxMessageBox(csz);
		return FALSE;
	}

	m_pBits = ((char *)(m_pFile)) + m_pBMFH->bfOffBits;

	if (m_pMonochromeBitmap) {
		delete m_pMonochromeBitmap;
		m_pMonochromeBitmap = NULL;
	}

	m_bThresholdTable.SetSize(0);

	switch(m_pBMI->bmiHeader.biBitCount) {
	default: {
		CString csz;
		AfxFormatString1(csz,IDS_FILE_FORMAT_NOT_SUPPORTED,lpszPathName);
		return FALSE;
			 }

	case 1:
		m_nRasterWidthInBytes = (m_pBMI->bmiHeader.biWidth + 7) / 8;
		m_pMonochromeBitmap = new CDIB1MonochromeBitmap;
		break;
	case 4:
		m_nRasterWidthInBytes = (m_pBMI->bmiHeader.biWidth + 1) / 2;
		m_pMonochromeBitmap = new CDIB4MonochromeBitmap;
		if (m_pBMI->bmiHeader.biClrUsed == 0) {
			m_bThresholdTable.SetSize(16);
		} else {
			m_bThresholdTable.SetSize(m_pBMI->bmiHeader.biClrUsed);
		}
		break;
	case 8:
		m_nRasterWidthInBytes = m_pBMI->bmiHeader.biWidth;
		m_pMonochromeBitmap = new CDIB8MonochromeBitmap;
		if (m_pBMI->bmiHeader.biClrUsed == 0) {
			m_bThresholdTable.SetSize(256);
		} else {
			m_bThresholdTable.SetSize(m_pBMI->bmiHeader.biClrUsed);
		}
		break;
#ifdef NOTYET
	case 16:
		m_nRasterWidthInBytes = m_pBMI->bmiHeader.biWidth * 2;
		break;
	case 24:
		m_nRasterWidthInBytes = m_pBMI->bmiHeader.biWidth * 3;
		break;
	case 32:
		m_nRasterWidthInBytes = m_pBMI->bmiHeader.biWidth * 4;
		break;
#endif
	}
	m_pMonochromeBitmap->SetDoc(*this);
	m_nRasterWidthInBytes = ((m_nRasterWidthInBytes + 3) / 4) * 4;

	BuildThresholdTable();
	return TRUE;
}

//-----------------------------------------------------------------
//
// CDIBDoc::CloseFile - Close the file and back out its mapping.
//
//-----------------------------------------------------------------

void CDIBDoc::CloseFile()
{
	if (m_pFile) {
		ASSERT(m_hFileMapping);
		ASSERT(m_file.m_hFile != CFile::hFileNull);
		if (! UnmapViewOfFile(m_pFile)) {
			CFileException::ThrowOsError(GetLastError(),GetPathName());
		}
		m_pFile = NULL;
		m_hFileMapping = NULL;
	}
	m_file.Close();
}

//-----------------------------------------------------------------
//
// CDIBDoc::StretchBlt - Paint a device independent bitmap onto
//					     the given DC
//
// dc - paint onto this device context
// crDest - paint onto this rectangle
// crSrc - paint only the portion in this rect on the DIB
// dwROP - raster operation to use (see BitBlt, for instance)
//
//-----------------------------------------------------------------

void CDIBDoc::StretchBlt(CDC &dc,CRect &crDest, CRect &crSrc,DWORD dwROP)
{
	if (m_file.m_hFile == CFile::hFileNull) return;
	int nBitmapTop = Height() - crSrc.bottom;
	if (::StretchDIBits(dc.m_hDC,
					    crDest.left,
					    crDest.top,
					    crDest.Width(),
					    crDest.Height(),
						crSrc.left,
						nBitmapTop,
						crSrc.Width(),
						crSrc.Height(),
						m_pBits,
						m_pBMI,
						DIB_RGB_COLORS,
						dwROP) == GDI_ERROR) {
		CFileException::ThrowOsError(GetLastError(),NULL);
	}
}

//-----------------------------------------------------------------
//
// CDIBDoc::GetBitmapInfo - return the bitmap info structure for
//							the DIB.
//
//-----------------------------------------------------------------

BITMAPINFO * CDIBDoc::GetBitmapInfo()
{
	ASSERT(m_pBMI);
	return m_pBMI;
}

//-----------------------------------------------------------------
//
// CDIBDoc::GetBitmapBits - Return the pointer to the actual bitmap
//							data.
//
//-----------------------------------------------------------------

LPVOID CDIBDoc::GetBitmapBits()
{
	ASSERT(m_pBits);
	return m_pBits;
}

//-----------------------------------------------------------------
//
// CDIBDoc::GetRasterWidthInBytes - Return the number of bytes
//									needed to map one raster of data.
//
//-----------------------------------------------------------------

int CDIBDoc::GetRasterWidthInBytes()
{
	return m_nRasterWidthInBytes;
}

//-----------------------------------------------------------------
//
// CDIBDoc::GetThresholdTable - The threshold table contains one
//					boolean (TRUE for above threshold/FALSE for below)
//					for each color in the bitmap.
//
//-----------------------------------------------------------------

const CByteArray & CDIBDoc::GetThresholdTable()
{
	return m_bThresholdTable;
}

//-----------------------------------------------------------------
//
// CDIBDoc::GetMonochromeBitmap - return a reference to the monochrome
//					bitmap (used to generically hack an image)
//					that we built for this DIB.
//
//-----------------------------------------------------------------

CMonochromeBitmap &CDIBDoc::GetMonochromeBitmap()
{
	ASSERT(m_pMonochromeBitmap);
	return *m_pMonochromeBitmap;
}

//-----------------------------------------------------------------
//
// BYTE * CDIBDoc - Return a pointer to a raster of bitmap data. Note
//					that these will be in different formats for
//					different bitmap # of bits.
//
// nLine - Y coordinate of raster to fetch.
//-----------------------------------------------------------------

const BYTE * CDIBDoc::GetRaster(int nLine)
{
	ASSERT(m_pBits);
	ASSERT(m_pBMI);
	return ((const BYTE *)m_pBits) + (m_pBMI->bmiHeader.biHeight - nLine - 1) * m_nRasterWidthInBytes;
}

//-----------------------------------------------------------------
//
// CDIBDoc::Width - # of pixels per raster.
//
//-----------------------------------------------------------------

int CDIBDoc::Width()
{
	if (IsEmpty()) return 0;
	return m_pBMI->bmiHeader.biWidth;
}

//-----------------------------------------------------------------
//
// CDIBDoc::Height - # of rasters in bitmap
//
//-----------------------------------------------------------------

int CDIBDoc::Height()
{
	if (IsEmpty()) return 0;
	return m_pBMI->bmiHeader.biHeight;

}

//-----------------------------------------------------------------
//
// CDIBDoc::IsEmpty - returns TRUE if this is the null document.
//
//-----------------------------------------------------------------

BOOL CDIBDoc::IsEmpty()
{
    return (m_file.m_hFile == CFile::hFileNull);
}


//-----------------------------------------------------------------
//
// CDIBDoc::OnImageHistogram - Display a histogram indicating how
//							   many pixels are at each brightness.
//							   The dialog box allows the user to
//							   eyeball the histogram and set the
//							   threshold.
//
//-----------------------------------------------------------------

void CDIBDoc::OnImageHistogram() 
{
	CHistogramDlg dlg;
	vector<int> vBrightness;
	GetMonochromeBitmap().GetBrightnessHistogram(dlg.m_wndHistogram.m_vHistogram,vBrightness);
	for (int i = 0; i < vBrightness.size() && vBrightness[i] < m_nThreshold; i++);

	dlg.m_wndHistogram.m_nBound = i;

	if (dlg.DoModal() == IDOK) {
		m_nThreshold = vBrightness[dlg.m_wndHistogram.m_nBound];
		BuildThresholdTable();
		m_pMonochromeBitmap->ClearBlobList();
		m_pMonochromeBitmap->BuildBlobList();
		UpdateAllViews(NULL);
	}
}

//-----------------------------------------------------------------
//
// CDIBDoc::OnUpdateImageHistogram - Only allow this for non-monochrome
//								bitmaps.
//
//-----------------------------------------------------------------

void CDIBDoc::OnUpdateImageHistogram(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable((! IsEmpty()) && m_pBMI->bmiHeader.biBitCount > 1);	
}

//-----------------------------------------------------------------
//
// CDIBDoc::BuildThresholdTable - Build the boolean table that tells
//								whether a color is above or below
//								threshold.
//
//-----------------------------------------------------------------

void CDIBDoc::BuildThresholdTable()
{
	const RGBQUAD *pRGBQ = m_pBMI->bmiColors;
	for (int i = 0; i < m_bThresholdTable.GetSize(); i++) {
		switch(m_eThresholdMethod) {
		case eBlackBackground:
			if (pRGBQ[i].rgbRed + pRGBQ[i].rgbGreen + pRGBQ[i].rgbBlue > m_nThreshold) {
				m_bThresholdTable[i] =  TRUE;
			} else {
				m_bThresholdTable[i] = FALSE;
			}
			break;
		case eWhiteBackground:
			if (pRGBQ[i].rgbRed + pRGBQ[i].rgbGreen + pRGBQ[i].rgbBlue > m_nThreshold) {
				m_bThresholdTable[i] =  FALSE;
			} else {
				m_bThresholdTable[i] = TRUE;
			}
			break;
		}
	}
}
