//  SharpenImageFilter. 
import java.awt.image.*;

public class SharpenImageFilter extends ImageFilter {
    // these store the dimensions of the current image...
    protected int width, height;

    // these store the color components of each pixel,
    // we can use short array to save space since each component
    // of the RGBDefault color model must be between 0 and 255...
    protected short[] red = null;
    protected short[] green = null;
    protected short[] blue = null;
    protected short[] alpha = null;

    public void setDimensions( int width, int height) {
        this.width = width;
        this.height = height;
        super.setDimensions( width, height);
    }

    protected void allocateBuffersIfNecessary() {
        // allocate a buffer for each color component...
        if (red == null) {
            red = new short[width * height];
        }
        if (green == null) {
            green = new short[width * height];
        }
        if (blue == null) {
            blue = new short[width * height];
        }
        if (alpha == null) {
            alpha = new short[width * height];
        }
    }

    public void setPixels( int x, int y, int w, int h, ColorModel cm, int[] pixels, int offset, int scansize) {
        // make sure that we have buffers for each color component...
        allocateBuffersIfNecessary();

        // loop through all the samples in the pixels buffer and place them
        // in the proper locations in the color component buffers...
        int srcpos = offset;
        int destpos = y*width + x;
        for (int j = 0; j < h; j++) {
            for (int i = 0; i < w; i++) {
                // use the color model parameter to extract each 
                // of the 4 components...                
                red[destpos + i] = (short)cm.getRed(pixels[srcpos + i]);
                green[destpos + i] = (short)cm.getGreen(pixels[srcpos + i]);
                blue[destpos + i] = (short)cm.getBlue(pixels[srcpos + i]);
                alpha[destpos + i] = (short)cm.getAlpha(pixels[srcpos + i]);
            }
            srcpos += scansize;
            destpos += width;
        }
    }

    public void setPixels( int x, int y, int w, int h, ColorModel cm, byte[] pixels, int offset, int scansize) {
        // make sure that we have buffers for each color component...
        allocateBuffersIfNecessary();

        // loop through all the samples in the pixels buffer and place them
        // in the proper locations in the color component buffers...
        int srcpos = offset;
        int destpos = y*width + x;
        for (int j = 0; j < h; j++) {
            for (int i = 0; i < w; i++) {
                // the pixel value must be converted to an integer
                // between 0 ang 255, we need to clear the upper
                // bits to zero because java integer types are signed
                // and the cast from a byte to an int will result in
                // negative numbers for byte values greater than or
                // equal to 0x80...
                int t = ((int)pixels[srcpos + i] & 0x000000FF);
                
                // use the color model parameter to extract each 
                // of the 4 components...                                
                red[destpos + i] = (short)cm.getRed(t);
                green[destpos + i] = (short)cm.getGreen(t);
                blue[destpos + i] = (short)cm.getBlue(t);
                alpha[destpos + i] = (short)cm.getAlpha(t);
            }
            srcpos += scansize;
            destpos += width;
        }
    }

    public void imageComplete( int status) {
        if ((status == STATICIMAGEDONE) || (status == SINGLEFRAMEDONE)) {
            // we have the complete image,            
            // process it into a new set of samples using the
            // RGB default direct color model...
            int[] newImage = processImage();
            
            // pass it on to the next consumer...            
            consumer.setPixels( 0, 0, width, height, ColorModel.getRGBdefault(), newImage, 0, width);
        }

        // let the super class have a chance to process the
        // imageComplete() status, this will pass the status
        // on to the next consumer...
        super.imageComplete(status);
    }

    protected int[] processImage() {
        // process the image pixels...
        int[] pixels = processPixels();

        // we don't need the color channel buffers any more...
        red = null;
        green = null;
        blue = null;
        alpha = null;

        // return the result...
        return pixels;
    }
    
    protected int[] processPixels() {
        // these will hold the results of the sharpening
        // of the individual color components...
        int r, g, b, a;

        // this array will hold the processed samples...
        int[] pixels = new int[width * height];
        
        // this will maintain the current index into the 
        // pixels buffer...
        int p = 0;

        // first row just gets written to the destination...
        for (int i = 0; i < width; i++) {
            r = red[p];
            g = green[p];
            b = blue[p];            
            a = alpha[p];
            
            // convert components to a 32 bit sample in the
            // default RGB direct color model...
            pixels[p++] = (a << 24) | (r << 16) | (g << 8) | b;
        }
        
        // middle rows...        
        for (int j = 1; j < height - 1; j++) { 
            // first pixel in a row just gets written to the destination...
            r = red[p];
            g = green[p];
            b = blue[p];            
            a = alpha[p];            
            pixels[p++] = (a << 24) | (r << 16) | (g << 8) | b;
                        
            // middle pixels in each row get sharpened...
            for (int i = 1; i < width - 1; i++) {
                // sharpen the r,g, and b color components,
                // leave the alpha as it is...
                r = sharpen( red, p, width);
                g = sharpen( green, p, width);
                b = sharpen( blue, p, width);                                    
                a = alpha[p];                
                
                // pack the color components back into a 32 bit
                // default direct color model sample...
                pixels[p++] = (a << 24) | (r << 16) | (g << 8) | b;
            }
            
            // last pixel in row just gets written to the destination...
            r = red[p];
            g = green[p];
            b = blue[p];            
            a = alpha[p];                        
            pixels[p++] = (a << 24) | (r << 16) | (g << 8) | b;            
        }             

        // last row just gets written to the destination...
        for (int i = 0; i < width; i++) {
            r = red[p];
            g = green[p];
            b = blue[p];    
            a = alpha[p];                        
            pixels[p++] = (a << 24) | (r << 16) | (g << 8) | b;
        }
        
        // return the result...
        return pixels;
    }            
    
    private final int sharpen( short[] samples, int index, int rowspan) {
        // return the results of applying a sharpening filter
        // to a color component of a pixel at the given index...
        //
        // the function uses the following convolution kernel:
        // -1/8 -1/8 -1/8
        // -1/8   2  -1/8
        // -1/8 -1/8 -1/8
           
        int v = (((int)samples[index] << 4)
                    - samples[index + 1]
                    - samples[index - 1]                
                    - samples[index - rowspan - 1]        
                    - samples[index - rowspan]   
                    - samples[index - rowspan + 1]
                    - samples[index + rowspan - 1]        
                    - samples[index + rowspan]       
                    - samples[index + rowspan + 1]) >> 3;  

        // clip the result to be between 0 and 255...
        if (v < 0)
            return 0;
        else if (v > 255)
            return 255;
        else
            return v;        
    }
}

